<?
/*
(C) 2004-2013 DatuX - info@datux.nl

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

	require_once ("../common.php");
	require_once ("firewall.php");
	require_once ("services.php");
	require_once ("ipblocks.php");
	//require_once ("../network/network.php"); //NIET DOEN, IVM 'restart' vanaf CLI!

	$IPTABLES="/usr/sbin/iptables";
	$ERRORFILE="/etc/firewall/errors";

	$ZONES=array(
		"zone_LAN"=>"eth0",
		"zone_INET"=>"eth1",
		"zone_DMZ"=>"eth2",
		"zone_extra1"=>"eth3",
		"zone_extra2"=>"eth4",
		"zone_extra3"=>"eth5",
		"zone_IPSEC"=>"ipsec0",
		"zone_PPTP"=>"ppp+",
	);

	$POSTUP_FIREWALL="/etc/firewall/firewall.up";


	function IpTablesReadErrors()
	{
		global $ERRORFILE;
		return(FileReadSettings($ERRORFILE));
	}

	function IpTablesReadError($errorid)
	{
		global $ERRORFILE;
		return(FileReadSetting($ERRORFILE,$errorid));
	}

	function IptablesRun($params,$errorid=false)
	{
		global $IPTABLES;
		global $ERRORFILE;

		$exitcode=Cmd("$IPTABLES $params","", $out, $err);
		//something when wrong?
		if ($exitcode)
		{
			//store/log errors?
			if ($errorid)
			{
				LogWarn(_LogFirewallError($err[0]));
				FileWriteSetting($ERRORFILE,$errorid,_LogFirewallError($err[0]));
			}
			return $err[0];
		}
		//remove error from errorfile
		if ($errorid)
		{
			FileWriteSetting($ERRORFILE,$errorid);
		}
		return false;
	}

	//converts a rulearray into iptables parameters
	//add=1 to create add parameters
	//add=0 to create delete parameters
	//-it returns an array of resulting iptable rules.
	//-one rule can create multiple iptable rules. (happens with 'pat')
	function Rule2Params($add,$rule)
	{
		global $IPBLOCKS;
		global $ZONES;

		$param="";

		// ////////////////// TABLE
		if ($rule[access]=="pat")
		{
			$param.="-t nat ";
		}
		else
		{
			$param.="-t filter ";
		}

		// /////////// ADD OR DELETE RULE?
		if ($add)
		{
			//add rule
			$adddel="-A";
		}
		else
		{
			//delete rule
			$adddel="-D";
		}
		$param.="$adddel";

		// ////////////////// CHAIN
		if ($rule[access]=="pat")
		{
			//pat
			$param.="PREROUTING";
		}
		elseif ($rule[srctype]=="zone_FW")
		{
			//output chain
			$param.="output_".$rule[access];
		}
		elseif ($rule[dsttype]=="zone_FW")
		{
			//input chain
			$param.="input_".$rule[access];
		}
		else
		{
			//forward chain
			$param.="forward_".$rule[access];
		}

		// //////////////SOURCE ADDRESS / INTERFACE
		//is it a ipblock name?
		if ($rule[srctype]=="ipblock")
		{
			//get the iprange
			$src=" -s ".$IPBLOCKS[$rule[src]];
		}
		//host or ip adress
		elseif ($rule[srctype]=="hostname" || $rule[srctype]=="ipadress")
		{
			//use specified host
			$src=" -s ".$rule[src];
		}
		//bridge port
		elseif ($rule[srctype]=="bridgeport")
		{
			$src=" -m physdev --physdev-in ".$rule[src];
		}
		elseif($rule[srctype]=="mac_adres")
		{
			$src=" -m mac --mac-source ".$rule[src];
		}
		//network interface
		elseif ($ZONES[$rule[srctype]])
		{
			if ($rule[srctype]=="zone_IPSEC")
			{
				//ipsec changed from klips to netkey so we dont have a ipsec0 device anymore.
				//for now we implement it this way to stay compatible with the current syn-3 firewalling rules:
				$src=" -m policy --pol ipsec --dir in";
			}
			else
			{
				$src=" -m policy --pol none --dir in -i ".$ZONES[$rule[srctype]];
			}
		}
		//openvpn certificate
		elseif ($rule[srctype]=="openvpn_cert")
		{
			$src=" -m set --match-set openvpn_cert_".$rule[src]." src";
		}
		//this server
		else
		{
			$src="";
		}
		$param.=$src;

		// //////////////DST ADDRESS / INTERFACE
		//is it a ipblock name?
		if ($rule[dsttype]=="ipblock")
		{
			//get the iprange
			$param.=" -d ".$IPBLOCKS[$rule[dst]];
		}
		//host or ip adress
		elseif ($rule[dsttype]=="hostname" || $rule[dsttype]=="ipadress")
		{
			//use specified host
			$param.=" -d ".$rule[dst];
		}
		//bridge port
		elseif ($rule[dsttype]=="bridgeport")
		{
			//specify -m physdev only once
			if ($rule[srctype]!="bridgeport")
				$param.=" -m physdev";

			//CAUTION: matching on bridge output port works only with bridged packets!
			$param.=" --physdev-is-bridged --physdev-out ".$rule[dst];
		}
		//network interface
		elseif ($ZONES[$rule[dsttype]])
		{
			if ($rule[dsttype]=="zone_IPSEC")
			{
				//ipsec changed from klips to netkey so we dont have a ipsec0 device anymore.
				//for now we implement it this way to stay compatible with the current syn-3 firewalling rules:
				$param.=" -m policy --pol ipsec --dir out";
			}
			else
			{
				$param.=" -m policy --pol none --dir out -o ".$ZONES[$rule[dsttype]];
			}
		}
		//openvpn certificate
		elseif ($rule[dsttype]=="openvpn_cert")
		{
			$dst=" -m set --match-set openvpn_cert_".$rule[dst]." dst";
		}
		//this server
		else
		{
		}

		// //////////////SERVICE DST PORT OR PROTOCOL
		if (($rule[dstservicetype]=='service'))
		{
			global $SERVICES;
			$service=$SERVICES[$rule[dstservice]];
			$dstservice="";
			$dstproto="";
			if ($service[proto])
			{
				$dstservice=" -p $service[proto]";
				$dstproto=" -p $service[proto]";
				//port specified?
				if ($service[port]!='')
				{
					$dstservice.=" --dport $service[port]";
					//port range specified?
					if ($service[toport]!='')
					{
						$dstservice.=":$service[toport]";
					}
				}
			}

			$param.=$dstservice;
		}

		// //////////////TARGET
		if ($rule[dstservicetype]=='group')
		{
			//for service group-matching we jump to a service-group target:
			//replace spaces by underscores
			$chain=str_replace(" ","_",$rule[dstservice]);
			if ($rule[access]=='allow')
			{
				//jump to the Accept-chain of the service group
				$target="A'$chain'";
			}
			elseif ($rule[access]=='deny')
			{
				//jump to the Drop-chain of the service group
				$target="D'$chain'";
			}
		}
		elseif ($rule[dstservicetype]=='service')
		{
			//for normal services (non-grouped) we use a standard iptables target:
			$targets[allow]="ACCEPT";
			$targets[deny]="DROP";
			$targets[pat]="DNAT --to ".$rule[fwdhost].":".$rule[fwdport];
			$target=$targets[$rule[access]];
		}
		$param.=" -j $target";

		// //////////////CREATE EXTRA FORWARD RULE FOR PAT?
		if ($rule[access]=="pat")
		{
			//pat is always from Zone Internet and should be unencrypted traffic
			$param.="  -m policy --pol none --dir in";

			//create extra FORWARD rule to allow this pat.
			//Here we need to ACCEPT the DNATtted packets
			list($port,$toport)=explode("-",$rule[fwdport]);
			if(isset($toport))
				$ports="$port:$toport";
			else
				$ports=$rule[fwdport];

			$extrarule="-t filter $adddel"."forward_pat $src -d $rule[fwdhost] $dstproto --dport $ports -j ACCEPT";
		}
		else
		{
			$extrarule="";
		}

		if ($extrarule)
			return array($param,$extrarule);
		else
			return array($param);
	}

	function IptablesChangeRule($oldrule,$newrule=false,$id)
	{
		//delete old rule?
		if ($oldrule[enabled])
		{
			$params=Rule2Params(0,$oldrule);
			foreach ($params as $param)
				IpTablesRun($param,$id);
		}

		//add new rule?
		if ($newrule[enabled])
		{
			$params=Rule2Params(1,$newrule);
			foreach ($params as $param)
			{
				$err=IpTablesRun($param,$id);
				//stop on the first error (otherwise the error could 'dissappear' if the next rule succeeds)
				if ($err)
				{
					return ($err);
				}
			}
		}
		return ($err);
	}


	//updates the list of administration ips in iptables
	function IptablesSetAdminIps($ips)
	{
		//flush adminips
		IptablesRun("-Fadmins",'flushadmins');
		//reset them
		foreach ($ips as $ip=>$bla)
		{
			IptablesRun("-Aadmins -s $ip -j ACCEPT",$ip);
		}

		//always allow access from/to these adresses, in case the sysad screwsup
/*		DISABLED - CUSTOMER SHOULD BE RESPONSIBLE FOR HIS OWN FIREWALL
		$datux_ips=array(
			"86.90.21.66", // edwin thuis
			"77.74.48.183", // lounge
			"193.239.186.219", //nieuw kantoor erwin thuis
		);
		foreach ($datux_ips as $ip)
		{
			IptablesRun("-Aadmins -s $ip -j ACCEPT",$ip);
		}*/
	}

	//creates serivcesgroup chains and adds services
	function IptablesAddServices()
	{
		global $SERVICES;
		foreach ($SERVICES as $id=>$service)
		{
			$group=$service['group'];
			//spaces should be converted to underscores!
			$group=str_replace(" ","_",$group);
			if ($group!='')
			{
				//still need to create chains?
				if (!$chainscreated[$group])
				{
					//accept and drop chains
					IptablesRun("-N'A$group'");
					IptablesRun("-N'D$group'");
					$chainscreated[$group]=1;
				}


				if ($service[proto])
				{
					$dstservice=" -p $service[proto]";
					//port specified?
					if ($service[port]!='')
					{
						$dstservice.=" --dport $service[port]";
						//port range specified?
						if ($service[toport]!='')
						{
							$dstservice.=":$service[toport]";
						}
					}
				}
				IptablesRun("-A'A$group' $dstservice -j ACCEPT");
				IptablesRun("-A'D$group' $dstservice -j DROP");
			}
		}
	}

	function IptablesRestart()
	{
		global $ZONES;

		LogInfo(_LogRestartingFirewall());

		UpdateOpenVPNIpsets();
		
		$rules=firewall_Read();

		//enable ip forwarding in kernel
		FileWriteRaw("/proc/sys/net/ipv4/ip_forward","1",0);

		//load certain modules
		Modprobe("ip_nat_irc");
		Modprobe("ip_nat_ftp");
		Modprobe("ip_nat_pptp");
		Modprobe("ip_nat_tftp");
		//Modprobe("ip_nat_snmp_basic");
		Modprobe("ip_nat_amanda");
		Modprobe("nf_conntrack_netbios_ns");
		//Geeft meer problemen met sip dan dat het oplost:
		//Modprobe("nf_nat_sip");


		//temporary input accept policy (otherwise drbd could deadlock and DNS gives trouble)
		//(will be reset at the end of this script)
		IptablesRun("-PINPUT ACCEPT");

		//flush all rules and chains
		IptablesRun("-t nat -F");
		IptablesRun("-t nat -X");
		IptablesRun("-F");
		IptablesRun("-X");

		//temporary input accept rule (otherwise drbd could deadlock and DNS gives trouble)
		//(will be reset at the end of this script)
		IptablesRun("-IINPUT -j ACCEPT");

		//forward policies is drop
		//output is always allowed
		IptablesRun("-PFORWARD DROP");
		IptablesRun("-POUTPUT ACCEPT");

		//create chains
		IptablesRun("-Nforward_pat");
		IptablesRun("-Nforward_deny");
		IptablesRun("-Nforward_allow");
		IptablesRun("-Noutput_deny");
		IptablesRun("-Noutput_allow");
		IptablesRun("-Ninput_deny");
		IptablesRun("-Ninput_allow");
		IptablesRun("-Ncommon");

		//webint administrator ips
		//(special case because the firewall is configured via the webinterface)
		IptablesRun("-Nadmins");

		//from loopback and to loopback everything is allowed
		//IptablesRun("-AOUTPUT -o lo -j ACCEPT");
		IptablesRun("-AINPUT  -i lo -j ACCEPT");

		//the admin ip's are a special case
		//they ALWAYS have access to 10000 and 22, no matter what other rules
		//are applied, or how screwed up things are :D
		IptablesRun("-AINPUT -j admins -p tcp --dport 10000");
		IptablesRun("-AINPUT -j admins -p tcp --dport 22");
		//IptablesRun("-AOUTPUT -j admins -p tcp --sport 10000");
		//IptablesRun("-AOUTPUT -j admins -p tcp --sport 22");

		//packets that want to go into this server
		IptablesRun("-AINPUT -j input_deny");
		IptablesRun("-AINPUT -j common");
		IptablesRun("-AINPUT -m state --state NEW -j input_allow");

		//packets that originate from this server
		//output is always allowed
		//IptablesRun("-AOUTPUT -j ACCEPT");
		//IptablesRun("-AOUTPUT -j common");
		//IptablesRun("-AOUTPUT -j output_deny");
		//IptablesRun("-AOUTPUT -m state --state NEW -j output_allow");

		//packets that go _trough_ this server, from one zone to another
		//first pat is checked, then deny, then allow
		IptablesRun("-AFORWARD -j common");
		IptablesRun("-AFORWARD -m state --state NEW -j forward_pat");
		IptablesRun("-AFORWARD -j forward_deny");
		IptablesRun("-AFORWARD -m state --state NEW -j forward_allow");

		//common for all traffic
		//drop invalid shit
		IptablesRun("-Acommon -m state --state INVALID -j DROP");
		//allow eshtablised conns
		IptablesRun("-Acommon -m state --state ESTABLISHED,RELATED -j ACCEPT");

		//masquerading will only take place on internet zone for non-ipsec traffic
		IptablesRun("-t nat -APOSTROUTING -o $ZONES[zone_INET] -m policy --pol none --dir out -j MASQUERADE");

		//allow administrator ips
		global $ADMIN_IPS;
		IptablesSetAdminIps($ADMIN_IPS);

		//create services chains
		IptablesAddServices();

		//apply zah rules
		foreach ($rules as $ruleid=>$rule)
		{
			IptablesChangeRule(false,$rule,$ruleid);
		}

		//run custom script
		global $POSTUP_FIREWALL;
		$exitcode=Cmd("$POSTUP_FIREWALL 2>&1","", $out, $err);
		//something when wrong?
		if ($exitcode==1)
			LogWarn(_LogFirewallError(implode("\n",$out)));
		elseif ($exitcode==2)
			LogError(_LogFirewallError(implode("\n",$out)));

		//all the rules are in place, reset the policy to drop
		//and remove temporary accept rule
		IptablesRun("-PINPUT DROP");
		IptablesRun("-DINPUT -j ACCEPT");

		LogInfo(_LogFirewallRestartComplete());
	}

	if ($_SERVER[argv][1]=="firewall_restart")
	{
		IptablesRestart();

		if ($ADMIN_IPS)
		{
			echo _ServerReachableAt()."\n";
			foreach($ADMIN_IPS as $ip=>$bla)
			{
				echo "$ip\n";
			}
		}
		exit(0);
	}
